=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed by Fredo6 - Copyright April 2009

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   Lib6Palette.rb
# Original Date	:   8 May 2009 - version 1.0
# Description	:   Interactive button palette
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end


module Traductor

T6[:T_DLG_EdgePlain] = "Plain"
T6[:T_DLG_EdgeAll] = "ALL"

T6[:Palette_TIP_TipZone] = "Information area"
T6[:Palette_TIP_MsgZone] = "Message area"
T6[:Palette_TIP_Pos] = "Position Top or Bottom"
T6[:Palette_TIP_Shrink] = "Shrink / Expand Palette"

T6[:DEFAULT_SectionPalette] = "Palette Configuration"
T6[:DEFAULT_PalettePos] = "Palette in Top position"
T6[:DEFAULT_PaletteShrinked] = "Palette shrinked"

#--------------------------------------------------------------------------------------------------------------
# Button Palette
#--------------------------------------------------------------------------------------------------------------			 

class Palette

Palette_Button6 = Struct.new "Palette_Button6", :symb, :rank, :state, :tooltip, :tr, :tr1, :tr2, 
                                                :width, :height, :quad, :quad1, :quad2, :default_value, :edge_prop,
                                                :buttons, :value_proc, :action_proc, :draw_proc, :draw_code, :instructions,
                                                :bkcolor, :text, :xleft, :ybot, :computed, :draw_scale, :draw_refresh,
											    :parent, :value, :type, :multi, :main_color, :frame_color, 
											    :passive, :hidden, :grayed, :selected, :grayed_proc, :text_proc,
												:draw_extra_proc

attr_reader :top_pos, :shrinked

@@persistence = nil

#--------------------------------------
# Managing default parameters
#--------------------------------------

#Configure Palette Default Parameters
def Palette.config_default_parameters
	MYDEFPARAM.separator :DEFAULT_SectionPalette
	MYDEFPARAM.declare :DEFAULT_PalettePos, true, 'B'
	MYDEFPARAM.declare :DEFAULT_PaletteShrinked, false, 'B'
end

#Saving parameters across sessions
def persistence_save
	@@persistence = {} unless @@persistence

	#Common to all tools
	hsh = @@persistence
	hsh["PalettePos"] = @top_pos
	hsh["PaletteShrinked"] = @shrinked	
end

#Restoring parameters across sessions
def persistence_restore
	hsh = @@persistence
	unless hsh
		hsh = @@persistence = {}
		hsh["PalettePos"] = MYDEFPARAM[:DEFAULT_PalettePos]
		hsh["PaletteShrinked"] = MYDEFPARAM[:DEFAULT_PaletteShrinked]
	end
	@top_pos = hsh["PalettePos"]
	@shrinked = hsh["PaletteShrinked"]	
end

#Create a palette instance
def initialize(*args)
	#Default value
	@height = 32
	@top_pos = true
	@message = nil
	persistence_restore
		
	#parsing the arguments
	args.each do |arg|	
		arg.each { |key, value|  parse_palette_args(key, value) } if arg.class == Hash
	end
	
	#initialization
	@height2 = @height / 2
	
	@lst_symbs = nil
	@lst_buttons = []
	@hsh_buttons = {}
	@widmax_tip = 400
	@wid_ctrl = 16
	@lrectnum = [0, 0, 1, 1]
	@tra_basis = [1, 0, 0, 0] + [0, -1, 0, 0] + [0, 0, 1, 0] + [0, 0, 0, 1]
	@computed = false	
	@ptxy = Geom::Point3d.new 0, 0, 0
	@origin = Geom::Point3d.new 0, @height, 0
	
	@bkcolor_normal = 'lightgrey'
	@bkcolor_hilight = 'lightgrey'
	@bkcolor_selected = 'oldlace'
	@bkcolor_down = 'oldlace'
	@frcolor_white = 'white'
	@frcolor_gray = 'gray'
	@frcolor_dark = 'darkgray'
	
	@id_cursor_hand = MYPLUGIN.create_cursor "Cursor_Hand", 2, 2	
	
	@tooltip = ""
	@nb_sepa = 0
	
	@view = Sketchup.active_model.active_view
	
	#Open GL instance for drawing buttons
	@ogl = Traductor::OpenGL_6.new
	@standard_draw_proc = @ogl.method "draw_proc"
	
	#Create the standard buttons
	@button_shrink = create_button :_bt_shrink, { :tooltip => T6[:Palette_TIP_Shrink], :width => @height2, :rank => 1 }
	@button_pos = create_button :_bt_pos, { :tooltip => T6[:Palette_TIP_Pos], :width => @height2 }
	@button_tip = create_button :_bt_tip, { :tooltip => T6[:Palette_TIP_TipZone], :passive => true }
	@button_message = create_button :_bt_message, { :tooltip => T6[:Palette_TIP_MsgZone], :passive => true, 
	                                                :height => 16, :hidden => true }
	adjust_draw_standard_button
end

#Assign the individual propert for the palette
def parse_palette_args(key, value)
	skey = key.to_s
	case skey
	when /top_pos/i
		@top_pos = value
	when /shrinked/i
		@shrinked = value
	when /height/i
		@height = value
	when /proc/i
		@palette_proc = value
	end	
end

def [](symb)
	@hsh_buttons[symb]
end

def button_get_value(symb)
	button = @hsh_buttons[symb]
	(button) ? button.value : nil
end

#Create a button structure
def create_button(symb, *args, &action_proc)
	button = @hsh_buttons[symb]
	button = Palette_Button6.new unless button
	button.symb = symb
	button.action_proc = action_proc
	args.each do |arg|	
		arg.each { |key, value|  parse_button_args(button, key, value) } if arg.class == Hash
	end
	@hsh_buttons[symb] = button
	button
end

#Declare the parent of a button
def button_set_parent(button, sparent)
	parent = @hsh_buttons[sparent]
	parent = create_button sparent unless parent
	parent.buttons = [] unless parent.buttons
	parent.buttons.push button unless parent.buttons.include?(button)
	button.parent = parent
end

def button_set_args(symb, *args)
	button = @hsh_buttons[symb]
	return unless button
	args.each do |arg|	
		arg.each { |key, value|  parse_button_args(button, key, value) } if arg.class == Hash
	end
end

def button_set_draw_proc(symb, draw_proc, main_color=nil, frame_color=nil)
	button = @hsh_buttons[symb]
	return unless button
	parse_button_args button, :draw_proc, draw_proc
	parse_button_args button, :main_color, main_color if main_color
	parse_button_args button, :frame_color, frame_color if frame_color
	button.instructions = nil
end

def adjust_draw_standard_button
	button_set_draw_proc :_bt_shrink, ((@shrinked) ? :triangle_EP : :triangle_WP), "red"
	button_set_draw_proc :_bt_pos, ((@top_pos) ? :triangle_SP : :triangle_NP), "green"
end

def parse_button_args(button, key, value)
	skey = key.to_s
	case skey
	when /width/i
		button.width = value
	when /height/i
		button.height = value
	when /hidden/i
		button.hidden = value
	when /parent/i
		button_set_parent button, value
	when /bk_color/i
		button.bkcolor = value
	when /rank/i
		button.rank = value
	when /tip/i
		button.tooltip = value
	when /text_proc/i
		button.text_proc = value
	when /text/i
		button.text = value
	when /passive/i
		button.passive = value
	when /main_color/i
		button.main_color = value
	when /frame_color/i
		button.frame_color = value
	when /type/i
		button.type = value
	when /edge_prop/i
		button.edge_prop = value
	when /default_value/i
		button.default_value = value
	when /draw_scale/i
		button.draw_scale = value
	when /draw_refresh/i
		button.draw_refresh = value
	when /draw_extra_proc/i
		button.draw_extra_proc = value
	when /draw_proc/i
		if value.class == Symbol
			button.draw_proc = @standard_draw_proc
			button.draw_code = value
		else
			button.draw_proc = value
		end	
	when /value_proc/i
		button.value_proc = value
	when /grayed_proc/i
		button.grayed_proc = value
	when /value/i
		button.value = value
	end	
	button.computed = false
end

#Declare a button
def declare_button(symb, *args, &action_proc)
	button = create_button symb, *args, &action_proc	
	@lst_buttons.push button unless button.parent
	@computed = false
	symb
end

def declare_separator
	s = '__sepa__' + @nb_sepa.to_s
	@nb_sepa += 1
	symb = s.intern
	declare_button symb, { :passive => true, :height => @height, :width => @height / 3, :draw_proc => :separator }
	symb
end

def concat_symb(symb, text)
	s = symb.to_s + '___' + text
	s.intern
end

def declare_edge_prop(symb, rgshow, *args, &action_proc)
	#declare_button(symb, { :type => 'multi' }, *args) { stateproc.call if stateproc }
	declare_button(symb, { :type => 'multi' }, *args, &action_proc)
	button = @hsh_buttons[symb]
	hp = { :parent => symb, :width => 30, :main_color => button.main_color }
	button.edge_prop = 'PSMH' unless button.edge_prop
	list = []
	list.push ['P', T6[:T_DLG_EdgePlain], :edge_prop_plain] if button.edge_prop =~ /P/i
	list.push ['S', T6[:T_DLG_EdgeSoft], :edge_prop_soft] if button.edge_prop =~ /S/i
	list.push ['M', T6[:T_DLG_EdgeSmooth], :edge_prop_smooth] if button.edge_prop =~ /M/i
	list.push ['H', T6[:T_DLG_EdgeHidden], :edge_prop_hidden] if button.edge_prop =~ /H/i
	list.each do |ll|
		code = ll[0]
		next unless !rgshow || code =~ rgshow
		tip = ll[1]
		sdraw = ll[2]
		declare_button concat_symb(symb, code), hp, { :draw_proc => sdraw, :value => code, :tooltip => tip }
	end	
	symb
end

def compute_button(button, forced=false)
	return if button.computed && !forced
	return compute_multi(button, forced=false) if button.type =~ /multi/i
	button.rank = 0 unless button.rank
	button.width = button.height unless button.width
	button.width = @height unless button.width
	button.height = button.width unless button.height
	button_make_quads button
	button.computed = true
end

def compute_multi(button, forced=false)
	return if button.computed && !forced
	w = 0
	button.height = @height2
	button.buttons.each do |b|
		b.rank = 0
		b.height = @height - button.height
		b.computed = false
		compute_button b
		w += b.width unless b.hidden	
	end
	button.width = w
	button.rank = 1
	button_make_quads button
	button.multi = true
	button.computed = true
end

def button_make_quads(button)
	button.quad = make_quad button.width, button.height
	button.quad1 = make_quad button.width, button.height, 1
	button.quad2 = make_quad button.width, button.height, 2
end

def check_multi_state(button)
	sval = button.value_proc.call if button.value_proc
	button.value = sval
	lv = (sval) ? sval.upcase.split(';;') : nil
	button.buttons.each do |b|
		if b.value
			b.selected = (lv && !lv.include?(b.value.upcase)) ? false : true
		else
			b.selected = false
		end	
	end	
end

def compute_all
	@hsh_buttons.each { |symb, button|	compute_button button } unless @computed
	
	@lst_buttons.each do |button|
		if button.multi
			check_multi_state button if button.multi
		elsif button.value_proc
			button.selected = button.value_proc.call
		end	
	end
	
	return if @computed
	
	#Compute all buttons
	
	#initialization
	vpwid = @view.vpwidth
	@yref = (@top_pos) ? @height + 1 : @view.vpheight - 1
	
	#compute the width for the buttons
	widbut = 0
	@lst_buttons.each do |button|
		widbut += button.width if (!button.type && button.rank == 0) || button.type =~ /multi/i
	end
	w = vpwid - widbut - @wid_ctrl
	@wid_tip = (w > @widmax_tip) ? @widmax_tip : w
	@xstart_button = @wid_tip + @wid_ctrl
	@button_tip.width = @wid_tip
	@button_message.width = @wid_tip + @wid_ctrl + widbut
	button_make_quads @button_tip
	button_make_quads @button_message
	
	#computing the basic buttons
	button_transfo @button_shrink, 0, 1
	button_transfo @button_pos, 0, 0
	button_transfo @button_tip, @wid_ctrl, 0
	message_transfo
	
	#Computing the other buttons
	xpos = @xstart_button
	@lst_buttons.each do |button|
		if button.type == nil
			button_transfo button, xpos, button.rank
			xpos += button.width if button.rank == 0
		elsif button.type =~ /multi/i
			button_transfo button, xpos, button.rank
			x = xpos
			button.buttons.each do |b|
				next if b.hidden
				button_transfo b, x, 0
				x += b.width
			end
			xpos += button.width
		end	
	end
	
	#Computing the list of buttons
	@lst_visible_buttons = @hsh_buttons.values.sort { |b1, b2| b1.symb.to_s <=> b2.symb.to_s }
	
	#Declaring computing OK
	@computed = true
end

#Set the transformation for the button
def message_transfo
	button = @button_message
	button.xleft = @tra_basis[12] = 0
	button.ybot = @tra_basis[13] = @yref + ((@top_pos) ? button.height : -@height)
	button.tr = Geom::Transformation.new @tra_basis
	@tra_basis[12] += 1
	@tra_basis[13] -= 1
	button.tr1 = Geom::Transformation.new @tra_basis
	@tra_basis[12] += 1
	@tra_basis[13] -= 1
	button.tr2 = Geom::Transformation.new @tra_basis
end

#Set the transformation for the button
def button_transfo(button, x, rank)
	button.xleft = @tra_basis[12] = x
	button.ybot = @tra_basis[13] = @yref - rank * button.height
	button.tr = Geom::Transformation.new @tra_basis
	@tra_basis[12] += 1
	@tra_basis[13] -= 1
	button.tr1 = Geom::Transformation.new @tra_basis
	@tra_basis[12] += 1
	@tra_basis[13] -= 1
	button.tr2 = Geom::Transformation.new @tra_basis
end

def make_quad(w, h, incr=0)
	pt1 = Geom::Point3d.new incr, incr, 0
	pt2 = Geom::Point3d.new incr, h - incr, 0
	pt3 = Geom::Point3d.new w - incr, h - incr, 0
	pt4 = Geom::Point3d.new w - incr, incr, 0
	[pt1, pt2, pt3, pt4]
end

def draw_button(view, button)
	return if button.hidden
	compute_button button
	bt_color = (grayed?(button)) ? nil : button.bkcolor
	clickable = button_is_clickable?(button)
	
	#drawing the area
	if !clickable
		bkcolor = (bt_color) ? bt_color : @bkcolor_normal
		width = 1
		color1 = @frcolor_dark
		color2 = @frcolor_dark
	elsif button.selected
		bkcolor = @bkcolor_selected
		width = 2
		color1 = @frcolor_gray
		color2 = @frcolor_white
	elsif button == @sel_button	
		bkcolor = @bkcolor_down
		width = 2
		color1 = @frcolor_gray
		color2 = @frcolor_white
	elsif button == @hi_button
		bkcolor = (bt_color) ? bt_color : @bkcolor_hilight
		width = 2
		color1 = @frcolor_white
		color2 = @frcolor_gray
	else	
		bkcolor = (bt_color) ? bt_color : @bkcolor_normal
		width = 1
		color1 = 'darkgray'
		color2 = 'darkgray'
	end	
	
	if width == 1
		t = button.tr
		quad = button.quad
		pts = quad.collect { |pt| t * pt }
		pts1 = [pts[1], pts[2]]
		pts2 = [pts[0], pts[3]]
	else
		t = button.tr1
		quad = button.quad1
		pts = quad.collect { |pt| t * pt }
		pts1 = [pts[0], pts[1], pts[2]]
		pts2 = [pts[2], pts[3], pts[0]]
	end	

	#drawing the background
	view.drawing_color = bkcolor
	view.draw2d GL_QUADS, pts
	
	#Drawing the content"
	content = draw_button_content view, button
	
	#drawing the border
	view.line_width = width
	view.line_stipple = ""
	view.drawing_color = color1
	view.draw2d GL_LINE_STRIP, pts1
	view.drawing_color = color2
	view.draw2d GL_LINE_STRIP, pts2
	
	#Compute the text
	button.text = button.text_proc.call if button.text_proc
		
	#draw a border for groups
	if clickable && width == 1 && !content && button.text == nil
		view.drawing_color = 'silver'
		view.line_width = 1
		pts = button.quad2.collect { |pt| t * pt }
		view.draw2d GL_LINE_LOOP, pts
	end	
	
	#Draw the text if any
	if button.text && button.text.length > 0
		view.draw_text t * Geom::Point3d.new(3, button.height+1, 0), button.text
	end
end

def draw_button_content(view, button)
	#button colors
	if grayed?(button)
		main_color = @frcolor_white
		frame_color = @frcolor_dark
	else
		main_color = button.main_color
		frame_color = button.frame_color
	end
		
	#getting the instructions
	unless button.instructions && !button.draw_refresh	
		if button.draw_code
			button.instructions = @standard_draw_proc.call(button.draw_code, button.width - 4, 
			                                               button.height - 4, main_color, frame_color,
														   button.draw_scale, button.draw_extra_proc)
		else
			button.instructions = button.draw_proc.call(button.symb, button.width - 4, button.height - 4) if button.draw_proc
		end	
	end
	
	return false unless button.instructions
	
	#processing the instructions
	@ogl.process_draw_GL view, button.tr2, button.instructions
end

#Top drawing method for the palette
def draw(view)
	compute_all
		
	#Drawing the elements
	list_visible.each { |button| draw_button view, button }
	
end

#Set the tooltip
def set_tooltip(tooltip=nil, level=nil)
	text = (tooltip) ? tooltip : ""
	set_message_and_tip @button_tip, text, level
end

#Set a Message
def set_message(message=nil, level=nil)
	text = (message) ? message : nil
	set_message_and_tip @button_message, message, level
	@button_message.hidden = (text) ? false : true
end

#Utility to set Message and Tip
def set_message_and_tip(button, text=nil, level=nil)
	button.text = text
	case level
	when /W/i
		color = 'yellow'
	when /I/i
		color = 'lightgreen'
	when /E/
		color = 'lightcoral'
	else
		color = nil
	end
	button.bkcolor = color
end

def toggle_shrinked
	@shrinked = !@shrinked
	adjust_draw_standard_button
	persistence_save
	onMouseMove_zero
	@view.invalidate
end

def toggle_top_down
	@top_pos = !@top_pos
	adjust_draw_standard_button
	persistence_save
	@computed = false
	compute_all
	onMouseMove_zero
	@view.invalidate
end

#Test the status Grayed of a button
def grayed?(button)
	return false unless button.grayed_proc
	new_status = button.grayed_proc.call
	if new_status != button.grayed
		button.grayed = new_status
		button.instructions = nil
	end
	new_status
end

#check if a button is not clickable
def button_is_clickable?(button)
	#!(button.passive || button.grayed || button.hidden)
	!(button.passive || grayed?(button) || button.hidden)
end

#Return the list of visible buttons
def list_visible
	if (@shrinked)
		ls = [@button_pos, @button_shrink]
	else
		compute_all unless @lst_visible_buttons
		ls = @lst_visible_buttons
	end
	ls
end

def locate_button(x, y)
	#Quick test if within the palette
	@mouse_button = nil
	
	#Finding the button
	list_visible.each do |button|
		if x >= button.xleft && x <= button.xleft + button.width && y <= button.ybot && y >= button.ybot - button.height
			@mouse_button = button
			break
		end	
	end
	@mouse_button
end

#Set the cursor
def onSetCursor
	return false unless @mouse_button
	if button_is_clickable?(@mouse_button)
		ic = @id_cursor_hand
	else
		ic = 0
	end
	UI::set_cursor ic
	ic
end

#Mouse move methods
def onMouseMove_zero
	onMouseMove -1, @x, @y, @view if @x
end

def onMouseMove(flags, x, y, view)
	return false if cancel_button_down(flags)	
	@x = x
	@y = y
	button = locate_button x, y
	unless button || @button_down
		@hi_button = nil
		return false
	end	
	
	if @button_down
		@hi_button = nil
		if button == nil || (@sel_button && button != @sel_button)
			return true
		end
	else	
		@hi_button = button
	end	
	if @hi_button
		@view.tooltip = @hi_button.tooltip
		@view.invalidate
	end	
	true
end

#Mouse enters the view port
def onMouseLeave(view)
	@hi_button = false
	@view.invalidate
end

#Mouse leaves the view port
def onMouseEnter(view)
end

#Cancel a button down event when Up event is not in the same area
def cancel_button_down(flags)
	if (flags >= 0) && @button_down && (flags & 1 != 1)
		@hi_button = nil
		@sel_button = nil
		@button_down = false
		@view.invalidate
	end	
end	

#Button click - Means that we end the selection
def onLButtonDown(flags, x, y, view)
	button = locate_button x, y
	return false unless button
	@button_down = true
	@view.tooltip = ""
	@sel_button = button
	@sel_button = nil unless button_is_clickable?(@sel_button)
	@view.invalidate
	true
end

#Button click - Means that we end the selection
def onLButtonUp(flags, x, y, view)
	button = locate_button x, y
	
	if button && !@button_down
		@hi_button = nil
		@view.invalidate
		return true
	end
	
	#Not an event for the palette
	unless @button_down && button
		@hi_button = nil
		@sel_button = nil
		@view.invalidate
		return false
	end
	
	@button_down = false
	
	#Trapping the click if clic up is on the same button as click down
	if @sel_button && button == @sel_button 
		execute_button @sel_button if button_is_clickable?(@sel_button)
	else
		
	end

	@sel_button = nil
	@view.invalidate
	true
end

#Execution of action when clicking on button
def execute_button(button)
	#skip passive, hidden, grayed buttons
	return unless button && button_is_clickable?(button)
	
	symb = button.symb
	
	#Standard
	if symb == :_bt_pos
		return toggle_top_down
	elsif symb == :_bt_shrink
		return toggle_shrinked
	end
	
	#Checking multiple button
	parent = button.parent
	if parent && parent.multi
		execute_multi button, parent
		button = parent
	elsif button.multi
		lv = []
		button.value = button.default_value	
	end	
	
	#Calling the custom proc
	if button.action_proc
		button.action_proc.call
	elsif @palette_proc	
		@palette_proc.call :click, button.symb, button.value
	end	
end

#Execution method parent button controlling multi-choice buttons
def execute_multi(button, parent)
	sval = parent.value
	lv = (sval) ? sval.split(';;') : []
	if button.selected
		lv -= [button.value]
	else
		lv.push button.value unless lv.include?(button.value)
	end
	parent.value = lv.join ';;'
end
	
end	# class Palette

end	#module Traductor

